# pylint: disable=E1136
"""This module contains helper classes and methods to assist with the general
function of this bot.

MastodonAmnesia - deletes old Mastodon toots
Copyright (C) 2021, 2022  Mark S Burgunder

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import json
import logging
import os
import sys
from dataclasses import asdict
from dataclasses import dataclass
from typing import Any
from typing import Optional
from typing import Type
from typing import TypeVar

import aiohttp
from minimal_activitypub import __display_name__ as ma_log_name
from minimal_activitypub.client_2_server import ActivityPub
from minimal_activitypub.client_2_server import ActivityPubError
from rich import print as rprint
from rich.logging import RichHandler

from . import __display_name__

BC = TypeVar("BC", bound="BotConfig")
ConfigClass = TypeVar("ConfigClass", bound="Configuration")
MC = TypeVar("MC", bound="MastodonConfig")

logger = logging.getLogger(__display_name__)


@dataclass
class BotConfig:
    """Dataclass holding configuration values for general behaviour of
    MastodonAmnesia."""

    # pylint: disable=too-many-instance-attributes
    log_level: str
    delete_after: int
    skip_deleting_pinned: Optional[bool]
    skip_deleting_faved: Optional[bool]
    skip_deleting_bookmarked: Optional[bool]
    skip_deleting_poll: Optional[bool]
    skip_deleting_dm: Optional[bool]
    skip_deleting_media: Optional[bool]
    skip_deleting_faved_at_least: Optional[int]
    skip_deleting_boost_at_least: Optional[int]

    def __init__(  # noqa: max-complexity: 12
        self: BC, config: Optional[dict[str, Any]]
    ) -> None:
        # pylint: disable=too-many-statements
        # pylint: disable=too-many-branches
        # It's not that bad ;)
        if not config:
            config = {}

        self.log_level = config.get("log_level", "WARN")

        std_out_formatter = logging.Formatter(
            "%(name)s[%(process)d] %(levelname)s %(message)s",
            datefmt="%H:%M:%S",
        )
        std_out_handler = RichHandler()
        std_out_handler.setFormatter(std_out_formatter)
        std_out_handler.setLevel(self.log_level)
        logger.addHandler(std_out_handler)

        self.delete_after = config.get("delete_after", None)
        self.skip_deleting_bookmarked = config.get("skip_deleting_bookmarked")
        self.skip_deleting_faved = config.get("skip_deleting_faved")
        self.skip_deleting_pinned = config.get("skip_deleting_pinned")
        self.skip_deleting_poll = config.get("skip_deleting_poll")
        self.skip_deleting_dm = config.get("skip_deleting_dm")
        self.skip_deleting_media = config.get("skip_deleting_media")
        self.skip_deleting_faved_at_least = config.get("skip_deleting_faved_at_least")
        self.skip_deleting_boost_at_least = config.get("skip_deleting_boost_at_least")

        if not self.delete_after:
            self._get_delete_after()

        if self.skip_deleting_bookmarked is None:
            self._get_skip_bookmarked()

        if self.skip_deleting_faved is None:
            self._get_skip_faved()

        if self.skip_deleting_pinned is None:
            self._get_skip_pinned()

        if self.skip_deleting_poll is None:
            self._get_skip_poll()

        if self.skip_deleting_dm is None:
            self._get_skip_dm()

        if self.skip_deleting_media is None:
            self._get_skip_media()

        if self.skip_deleting_faved_at_least is None:
            self._get_skip_faved_at_least()

        if self.skip_deleting_boost_at_least is None:
            self._get_skip_boost_at_lease()

    def _get_skip_poll(self: BC) -> None:
        """'Private' method to get skip deleting polls value from user if this
        value has not yet been configured."""
        rprint("Should polls be deleted when they get old enough?")
        y_or_n = input("[..] Please enter Y for yes or N for no: ")
        if y_or_n in ("Y", "y"):
            self.skip_deleting_poll = False
        elif y_or_n in ("N", "n"):
            self.skip_deleting_poll = True
        else:
            rprint("! ERROR ... please only respond with 'Y' or 'N'")
            rprint("! Cannot continue. Exiting now.")
            sys.exit(1)

    def _get_skip_dm(self: BC) -> None:
        """'Private' method to get skip deleting 'private' messages value from
        user if this value has not yet been configured."""
        rprint("Should Direct Messages be deleted when they get old enough?")
        y_or_n = input("[..] Please enter Y for yes or N for no: ")
        if y_or_n in ("Y", "y"):
            self.skip_deleting_dm = False
        elif y_or_n in ("N", "n"):
            self.skip_deleting_dm = True
        else:
            rprint("! ERROR ... please only respond with 'Y' or 'N'")
            rprint("! Cannot continue. Exiting now.")
            sys.exit(1)

    def _get_skip_media(self: BC) -> None:
        """'Private' method to get skip deleting toots with media value from
        user if this value has not yet been configured."""
        rprint(
            "Should Toots with attachements / pictures be deleted "
            "when they get old enough?"
        )
        y_or_n = input("[..] Please enter Y for yes or N for no: ")
        if y_or_n in ("Y", "y"):
            self.skip_deleting_media = False
        elif y_or_n in ("N", "n"):
            self.skip_deleting_media = True
        else:
            rprint("! ERROR ... please only respond with 'Y' or 'N'")
            rprint("! Cannot continue. Exiting now.")
            sys.exit(1)

    def _get_skip_faved_at_least(self: BC) -> None:
        """'Private' method to get skip deleting toots that have been
        favourited value from user if this value has not yet been
        configured."""
        rprint(
            "Should Toots being favourited a certain mimimum number of times be "
            "excluded from deletion even when they get old enough?"
        )
        rprint("(enter 0 to disregard this setting)")
        self.skip_deleting_faved_at_least = int(input("[..] Please enter number: "))

    def _get_skip_boost_at_lease(self: BC) -> None:
        """'Private' method to get skip deleting toots that have been boosted
        value from user if this value has not yet been configured."""
        rprint(
            "Should Toots being boosted a certain mimimum number of times be "
            "excluded from deletion even when they get old enough?"
        )
        rprint("(enter 0 to disregard this setting)")
        self.skip_deleting_boost_at_least = int(input("[..] Please enter number: "))

    def _get_skip_pinned(self: BC) -> None:
        """'Private' method to get skip deleting pinned toots value from user
        if this value has not yet been configured."""
        rprint("Should pinned toots be deleted when they get old enough?")
        y_or_n = input("[..] Please enter Y for yes or N for no: ")
        if y_or_n in ("Y", "y"):
            self.skip_deleting_pinned = False
        elif y_or_n in ("N", "n"):
            self.skip_deleting_pinned = True
        else:
            rprint("! ERROR ... please only respond with 'Y' or 'N'")
            rprint("! Cannot continue. Exiting now.")
            sys.exit(1)

    def _get_skip_faved(self: BC) -> None:
        """'Private' method to get skip deleting faveroited toots value from
        user if this value has not yet been configured."""
        rprint("Should favoured toots be deleted when they get old enough?")
        y_or_n = input("[..] Please enter Y for yes or N for no: ")
        if y_or_n in ("Y", "y"):
            self.skip_deleting_faved = False
        elif y_or_n in ("N", "n"):
            self.skip_deleting_faved = True
        else:
            rprint("! ERROR ... please only respond with 'Y' or 'N'")
            rprint("! Cannot continue. Exiting now.")
            sys.exit(1)

    def _get_skip_bookmarked(self: BC) -> None:
        """'Private' method to get skip deleting bookmarked toots from user if
        this value has not yet been configured."""
        rprint("Should bookmarked toots be deleted when they get old enough?")
        y_or_n = input("[..] Please enter Y for yes or N for no: ")
        if y_or_n in ("Y", "y"):
            self.skip_deleting_bookmarked = False
        elif y_or_n in ("N", "n"):
            self.skip_deleting_bookmarked = True
        else:
            rprint("! ERROR ... please only respond with 'Y' or 'N'")
            rprint("! Cannot continue. Exiting now.")
            sys.exit(1)

    def _get_delete_after(self: BC) -> None:
        """'Private' method to get delete after value from user if this value
        has not yet been configured."""
        rprint(
            "Please enter maximum age of retained toots in the "
            'format of "number unit"'
        )
        rprint('For example "1 weeks" or "3 days". Supported units are:')
        rprint(
            " - seconds\n"
            " - minutes\n"
            " - hours\n"
            " - days\n"
            " - weeks\n"
            " - months"
        )
        max_age = input("[..] Minimum age to delete toots (in seconds): ")
        max_age_parts = max_age.split(" ")
        max_age_number = int(max_age_parts[0])
        max_age_unit = max_age_parts[1]
        if max_age_unit == "seconds":
            self.delete_after = max_age_number
        elif max_age_unit == "minutes":
            self.delete_after = max_age_number * 3600
        elif max_age_unit == "hours":
            self.delete_after = max_age_number * 3600
        elif max_age_unit == "days":
            self.delete_after = max_age_number * 3600 * 24
        elif max_age_unit == "weeks":
            self.delete_after = max_age_number * 3600 * 24 * 7
        elif max_age_unit == "months":
            self.delete_after = max_age_number * 3600 * 24 * 30
        else:
            rprint("! Error ... unknown unit ({max_age_unit}) specified")
            rprint("! Cannot continue. Exiting now.")
            sys.exit(1)


@dataclass
class MastodonConfig:
    """Dataclass holding configuration values for Mastodon settings."""

    instance: str
    access_token: str

    @classmethod
    async def load_mastodon_config(
        cls: Type[MC], mastodon_ini: Optional[dict[str, Any]]
    ) -> MC:
        """Loads Mastodon configuration from file and returns a MastodonConfig
        class instance."""

        logger.debug(
            "MastodonConfig.load_mastodon_config - mastodon_ini: %s", mastodon_ini
        )

        if not mastodon_ini:
            mastodon_ini = {}

        base_url = mastodon_ini.get("base_url", None)
        instance = mastodon_ini.get("instance", None)
        if instance is None and base_url:
            instance = base_url

        if not instance:
            instance = input(
                "[..] Enter instance (domain name) for Mastodon account host: "
            )

        access_token = mastodon_ini.get("access_token", None)

        user_info = mastodon_ini.get("user_info", None)
        if not access_token and user_info:
            access_token = user_info.get("access_token", None)

        try:
            if not access_token:
                user_name = input("[..] Enter user for your Fediverse account: ")
                password = input("[..] Enter password for your Fediverse account: ")

                async with aiohttp.ClientSession() as session:
                    access_token = await ActivityPub.get_auth_token(
                        instance_url=instance,
                        username=user_name,
                        password=password,
                        session=session,
                    )

        except ActivityPubError as error:
            rprint(f"! Error when setting up Fediverse connection: {error}")
            rprint("! Cannot continue. Exiting now.")
            sys.exit(1)

        return cls(instance=instance, access_token=access_token)


@dataclass
class Configuration:
    """Dataclass to hold all settings for tootbot."""

    bot: BotConfig
    mastodon: MastodonConfig

    @classmethod
    async def load_config(
        cls: Type[ConfigClass], config_file_name: str = "config.json"
    ) -> ConfigClass:
        """Loads configuration from file and returns a Configuration class
        instance."""
        if os.path.exists(config_file_name):
            with open(file=config_file_name, encoding="UTF-8") as config_file:
                loaded_config = json.load(config_file)
        else:
            loaded_config = {}

        mastodon_ini = loaded_config.get("mastodon")
        if not mastodon_ini:
            mastodon_ini = loaded_config.get("Mastodon")
        mastodon_config = await MastodonConfig.load_mastodon_config(
            mastodon_ini=mastodon_ini
        )

        bot_ini = loaded_config.get("bot")
        if not bot_ini:
            bot_ini = loaded_config.get("Bot")
        bot = BotConfig(config=bot_ini)

        config_instance = cls(bot=bot, mastodon=mastodon_config)
        config_dict = asdict(config_instance)
        with open(file=config_file_name, mode="w", encoding="UTF-8") as config_file:
            json.dump(config_dict, config_file, indent=4)
            logger.debug("Configuration.load_config - Saved config: %s", config_dict)

        return config_instance


async def setup_shop(
    config_file: str,
    debug_log_file: Optional[str],
) -> Configuration:
    """Processes command line arguments, establishes debug logging to file if
    specified, loads configuration.

    Returns:
        bool: true if this is a dry-run, i.e. no toots should be actually deleted.
        Configuration: config for this run of the MastodonAmnesia.
    """
    file_log_formatter = logging.Formatter(
        "%(asctime)s %(name)s[%(process)d] %(levelname)s %(message)s",
        datefmt="%H:%M:%S",
    )
    if debug_log_file:
        file_handler = logging.FileHandler(filename=debug_log_file)
        file_handler.setFormatter(file_log_formatter)
        file_handler.setLevel(logging.DEBUG)
        logger.addHandler(file_handler)
        ma_logger = logging.getLogger(ma_log_name)
        ma_logger.setLevel(logging.DEBUG)
        ma_logger.addHandler(file_handler)

    config = await Configuration.load_config(config_file_name=config_file)
    return config



"""
    MastodonAmnesia - deletes old Mastodon toots
    Copyright (C) 2021, 2022  Mark S Burgunder

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import csv
import logging
import os
from dataclasses import dataclass
from enum import Enum
from typing import Any
from typing import Optional
from typing import Type
from typing import TypeVar

import aiocsv
import aiofiles
import arrow
from outdated import check_outdated
from rich import print  # pylint: disable=redefined-builtin

from . import __display_name__
from . import __package_name__
from . import __version__
from . import Toot

logger = logging.getLogger(__display_name__)


def check_updates() -> None:
    """Check if there is a newer version of MastodonAmnesia available on
    PyPI."""
    is_outdated = False
    try:
        is_outdated, pypi_version = check_outdated(
            package=__package_name__,
            version=__version__,
        )
        if is_outdated:
            print(
                f"[bold][red]!!! New version of {__display_name__} ({pypi_version}) "
                f"is available on PyPI.org !!!\n"
            )
            logger.debug(
                "check_updates - New version of %s (%s) is available on PyPI.org",
                __display_name__,
                pypi_version,
            )
    except ValueError:
        print(
            "[yellow]Notice - Your version is higher than last published version on PyPI"
        )
        logger.debug(
            "check_updates - Notice - Your version is higher than last published version on PyPI"
        )


AL = TypeVar("AL", bound="AuditLog")


@dataclass
class AuditLog:
    """Class to control the creation of an audit log and adding entries to
    it."""

    class Style(Enum):
        """Enumating different audit log file styles implemented."""

        PLAIN = 1
        CSV = 2

    audit_log: Any
    style: Style

    @classmethod
    async def open(
        cls: Type[AL],
        audit_log_file: str,
        style: Optional[str],
    ) -> AL:
        """Opens the audit log file for writing and appending new log entries
        and returns a new AuditLog instance."""
        audit_log = await aiofiles.open(file=audit_log_file, mode="at")
        if style:
            log_style = AuditLog.Style[style]
        else:
            log_style = AuditLog.Style.PLAIN

        if (
            log_style == AuditLog.Style.CSV
            and os.fstat(audit_log.fileno()).st_size == 0
        ):
            await AuditLog._add_csv_header(audit_log)

        return cls(audit_log=audit_log, style=log_style)

    async def close(self: AL) -> None:
        """Closes audit log file."""
        self.audit_log.close()

    async def add_entry(self: AL, toot: Toot) -> None:
        """Appends an entry/toot details to the audit log file."""

        if self.style == AuditLog.Style.PLAIN:
            await self._add_plain_entry(toot)
        elif self.style == AuditLog.Style.CSV:
            await self._add_csv_entry(toot)
        else:
            print(f"Unsupported log style passed: {self.style=}")

    async def _add_plain_entry(self: AL, toot: Toot) -> None:
        """Appends a plain text entry/toot details to the audit log file."""
        entry = (
            f"{arrow.now().format('YYYY-MM-DD hh:mm:ss')} -"
            f" Removed {'poll' if toot.get('poll', False) else 'status'} {toot.get('url')}"
            f" created @ {toot.get('created_at')}"
            f" with {toot.get('visibility')} visibility, {len(toot.get('media_attachments', []))} attachments."
            f" This toot was reblogged {toot.get('reblogs_count')} times and"
            f" favourited {toot.get('favourites_count')} times."
            f" The toot was {'pinned' if toot.get('pinned') else 'not pinned'}.\n"
        )
        await self.audit_log.write(entry)

    async def _add_csv_entry(self: AL, toot: Toot) -> None:
        """Appends a vsc formatted entry/toot details to the audit log file."""
        csv_writer = aiocsv.writers.AsyncWriter(
            asyncfile=self.audit_log,
            quoting=csv.QUOTE_ALL,
        )
        record = [
            arrow.now(),
            "poll" if toot.get("poll", False) else "status",
            toot.get("url"),
            arrow.get(toot.get("created_at", 0)).to(tz="local"),
            toot.get("visibility"),
            len(toot.get("media_attachments", [])),
            toot.get("reblogs_count"),
            toot.get("favourites_count"),
            toot.get("pinned"),
        ]
        await csv_writer.writerow(row=record)

    @staticmethod
    async def _add_csv_header(csv_file: Any) -> None:
        """Writes a CSV header line to the audit log file.

        This method is normally called when an audit log file is being
        initially created before writing any other entries to the audit
        log; i.e. The record it writes will be the header record for the
        CSV file.
        """
        csv_writer = aiocsv.writers.AsyncWriter(
            asyncfile=csv_file, quoting=csv.QUOTE_ALL
        )
        record = [
            "date and time of deletion",
            "type of toot,",
            "url of toot before deletion",
            "date and time deleted toot was created",
            "visibility",
            "# of media_attachments",
            "reblogs_count",
            "favourites_count",
            "pinned",
        ]
        await csv_writer.writerow(row=record)



"""
    MastodonAmnesia - deletes old Mastodon toots
    Copyright (C) 2021, 2022  Mark S Burgunder

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import asyncio
import json
import logging
import sys
from math import ceil
from typing import Any
from typing import cast
from typing import Optional

import aiohttp
import arrow
import click
from aiohttp.client_exceptions import ClientError
from minimal_activitypub.client_2_server import ActivityPub
from minimal_activitypub.client_2_server import ActivityPubError
from minimal_activitypub.client_2_server import RatelimitError
from rich import print  # pylint: disable=redefined-builtin
from rich import traceback
from tqdm import tqdm
from tqdm import trange

from . import __display_name__
from . import __version__
from . import Toot
from .config import Configuration
from .config import setup_shop
from .util import AuditLog
from .util import check_updates

traceback.install(show_locals=True)
logger = logging.getLogger(__display_name__)
logger.setLevel(logging.DEBUG)


async def main(  # noqa: max-complexity: 13
    config_file: str,
    is_dry_run: bool,
    debug_log_file: Optional[str],
    audit_log_file: Optional[str],
    audit_log_style: Optional[str],
    batch_size: Optional[int],
    limit: Optional[int],
) -> None:
    """Main logic to run MastodonAmnesia."""

    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-arguments

    config = await setup_shop(
        config_file=config_file,
        debug_log_file=debug_log_file,
    )

    oldest_to_keep = arrow.now().shift(seconds=-config.bot.delete_after)

    print(f"Welcome to {__display_name__} {__version__}")
    logger.debug("main -Welcome to %s %s", __display_name__, __version__)

    check_updates()

    try:
        session = aiohttp.ClientSession()

        instance = ActivityPub(
            instance=config.mastodon.instance,
            access_token=config.mastodon.access_token,
            session=session,
        )
        await instance.determine_instance_type()
        user_info = await instance.verify_credentials()
        print(
            f"We are removing toots older than {oldest_to_keep} "
            f"from {config.mastodon.instance}@{user_info['username']}"
        )

        audit_log = None
        if audit_log_file:
            print(
                f"A record of all deleted toots will be recorded in the audit log file at {audit_log_file}"
            )
            audit_log = await AuditLog.open(
                audit_log_file=audit_log_file,
                style=audit_log_style,
            )

        toots = await instance.get_account_statuses(account_id=user_info["id"])
    except RatelimitError:
        print(
            f"RateLimited during startup, [red]Please wait until[/red] "
            f"{instance.ratelimit_reset} before trying again"
        )
        sys.exit(429)
    except (ClientError, ActivityPubError):
        logger.exception("!!! Cannot continue.")
        sys.exit(100)

    toots_to_delete: list[Toot] = []
    title = "Finding toots to delete"
    progress_bar = tqdm(
        desc=f"{title:.<60}",
        ncols=120,
        unit="toots",
        position=0,
        bar_format="{l_bar} {n_fmt} at {rate_fmt}",
    )
    while True:
        try:
            for toot in toots:
                logger.debug(
                    "Processing toot: %s from %s",
                    toot.get("url"),
                    arrow.get(toot.get("created_at")).to(tz="local"),
                )

                logger.debug(
                    "Oldest to keep vs toot created at %s > %s",
                    oldest_to_keep,
                    arrow.get(toot.get("created_at")).to(tz="local"),
                )

                if should_keep(
                    toot=toot,
                    oldest_to_keep=oldest_to_keep,
                    config=config,
                ):
                    logger.info(
                        "Not deleting toot: "
                        "Bookmarked: %s - "
                        "My Fav: %s - "
                        "Pinned: %s - "
                        "Poll: %s - "
                        "Attachements: %s - "
                        "Faved: %s - "
                        "Boosted: %s - "
                        "DM: %s -+- "
                        "Created At: %s -+- "
                        "%s",
                        toot.get("bookmarked"),
                        toot.get("favourited"),
                        toot.get("pinned"),
                        (toot.get("poll") is not None),
                        len(toot.get("media_attachments")),
                        toot.get("favourites_count"),
                        toot.get("reblogs_count"),
                        (toot.get("visibility") == "direct"),
                        arrow.get(toot.get("created_at")).to(tz="local"),
                        toot.get("url"),
                    )

                elif limit and len(toots_to_delete) >= limit:
                    break

                else:
                    toots_to_delete.append(toot)

                progress_bar.update()

            if limit and len(toots_to_delete) >= limit:
                break

            # Get More toots if available:
            logger.debug("Main - get next batch of toots if available.")
            logger.debug("Main - instance.pagination: %s", instance.pagination)
            if (
                instance.pagination["next"]["max_id"]
                or instance.pagination["next"]["min_id"]
            ):
                toots = await instance.get_account_statuses(
                    account_id=user_info["id"],
                    max_id=instance.pagination["next"]["max_id"],
                    min_id=instance.pagination["next"]["min_id"],
                )
                logger.debug("Main - scrolling - len(toots): %s", len(toots))
                if len(toots) == 0:
                    break
            else:
                break

        except RatelimitError:
            await sleep_off_ratelimiting(instance=instance)

    progress_bar.close()

    total_toots_to_delete = len(toots_to_delete)
    logger.debug(
        "Main - start deleting - total_toots_to_delete: %s", total_toots_to_delete
    )
    for toot in toots_to_delete:
        logger.debug(
            "Start of deleting - toot to delete: %s @ %s",
            toot["id"],
            toot["url"],
        )

    # If dry-run has been specified, print out list of toots that would be deleted
    if is_dry_run:
        print("\n--dry-run or -d specified. [yellow][bold]No toots will be deleted")
        for toot in toots_to_delete:
            print(
                f"[red]Would[/red] delete toot"
                f" {toot.get('url')} from {toot.get('created_at')}"
            )
        print(f"Total of {total_toots_to_delete} toots would be deleted.")

    # Dry-run has not been specified... delete toots!
    else:
        await delete_toots(
            instance=instance,
            toots_to_delete=toots_to_delete,
            audit=audit_log,
            batch_size=batch_size,
        )
        print(f"All old toots deleted! Total of {total_toots_to_delete} toots deleted")

    if audit_log:
        await audit_log.close()
    await session.close()


async def delete_toots(
    instance: ActivityPub,
    toots_to_delete: list[Toot],
    audit: Optional[AuditLog],
    batch_size: Optional[int],
) -> None:
    """Method to delete all toots that should be deleted."""
    title = "Deleting toots"
    total_toots_to_delete = len(toots_to_delete)
    if total_toots_to_delete > 0:
        with tqdm(
            desc=f"{title:.<60}",
            ncols=120,
            total=total_toots_to_delete,
            unit="toots",
            position=0,
            bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt} at {rate_fmt}",
        ) as progress_bar:

            while len(toots_to_delete) > 0:

                tasks = []
                for toot in toots_to_delete:
                    tasks.append(
                        delete_single_toot(toot=toot, instance=instance, audit=audit)
                    )
                    if batch_size and len(tasks) >= batch_size:
                        break

                responses = await asyncio.gather(*tasks)

                # Filter out any responses that are None...
                # Those encountered Rate Limiting
                deleted_toots = [toot for toot in responses if toot is not None]
                logger.debug(
                    "delete_toots - len(deleted_toots): %s",
                    len(deleted_toots),
                )

                progress_bar.update(len(deleted_toots))

                if len(deleted_toots) < len(toots_to_delete):
                    await sleep_off_ratelimiting(instance=instance)

                for toot in deleted_toots:
                    toots_to_delete.remove(toot)


async def sleep_off_ratelimiting(
    instance: ActivityPub,
) -> None:
    """Determines time needed to wait and waits for rate limiting to be
    over."""

    logger.debug(
        "sleep_off_ratelimiting - Rate limited: Limit: %s - resetting at: %s",
        instance.ratelimit_remaining,
        instance.ratelimit_reset,
    )
    reset_at = arrow.get(instance.ratelimit_reset).datetime
    now = arrow.now().datetime
    need_to_wait = ceil((reset_at - now).total_seconds())

    logger.info(
        "Need to wait %s seconds (until %s) to let server 'cool down'",
        need_to_wait,
        arrow.get(instance.ratelimit_reset),
    )
    bar_title = "Waiting to let server 'cool-down'"
    for _i in trange(
        need_to_wait,
        desc=f"{bar_title:.<60}",
        unit="s",
        ncols=120,
        bar_format="{desc}: {percentage:3.0f}%|{bar}| Eta: {remaining} - Elapsed: {elapsed}",
        position=1,
    ):
        await asyncio.sleep(1)


async def delete_single_toot(
    toot: Toot,
    instance: ActivityPub,
    audit: Optional[AuditLog],
) -> Optional[dict[str, Any]]:
    """Deletes a single toot."""
    logger.debug(
        "delete_single_toot(toot=%s, instance=%s)",
        toot["id"],
        instance.instance,
    )
    return_toot: Optional[dict[str, Any]] = toot
    try:
        await instance.delete_status(status=toot)
        logger.info(
            "delete_single_toot - Deleted toot %s from %s",
            toot.get("url"),
            toot.get("created_at"),
        )
        if audit:
            await audit.add_entry(toot=toot)
    except RatelimitError:
        logger.debug(
            "delete_single_toot - toot id = %s - ratelimit_remaining = %s - ratelimit_reset = %s",
            toot["id"],
            instance.ratelimit_remaining,
            instance.ratelimit_reset,
        )
        return_toot = None
    except ActivityPubError as error:
        logger.debug(
            "delete_single_toot - encountered error: %s",
            error,
        )
        logger.debug("delete_single_toot - toot: %s", json.dumps(toot, indent=4))
        raise error

    return return_toot


def should_keep(toot: Toot, oldest_to_keep: arrow.Arrow, config: Configuration) -> bool:
    """Function to determine if toot should be kept even though it might be a
    candidate for deletion."""

    # Kind of agree, but alternative would be harder to read IMHO
    # pylint: disable=too-many-return-statements

    toot_created_at = arrow.get(cast(int, toot.get("created_at")))
    if toot_created_at >= oldest_to_keep:
        return True

    if config.bot.skip_deleting_bookmarked and bool(toot.get("bookmarked")):
        return True

    if config.bot.skip_deleting_faved and bool(toot.get("favourited")):
        return True

    if config.bot.skip_deleting_pinned and bool(toot.get("pinned")):
        return True

    if config.bot.skip_deleting_poll and bool(toot.get("poll")):
        return True

    if config.bot.skip_deleting_dm and toot.get("visibility") == "direct":
        return True

    medias = toot.get("media_attachments")
    if (
        config.bot.skip_deleting_media
        and isinstance(medias, list)
        and bool(len(medias))
    ):
        return True

    favourites = toot.get("favourites_count", 0)
    if (
        config.bot.skip_deleting_faved_at_least
        and favourites >= config.bot.skip_deleting_faved_at_least
    ):
        return True

    reblogs = toot.get("reblogs_count", 0)
    if (
        config.bot.skip_deleting_boost_at_least
        and reblogs >= config.bot.skip_deleting_boost_at_least
    ):
        return True

    return False


@click.command()
@click.option(
    "-c",
    "--config-file",
    help="Name of configuration file to use",
    default="config.json",
    type=click.Path(
        exists=True,
    ),
)
@click.option(
    "-d",
    "--dry-run",
    help="Only print out which toots would be deleted. No actual deletion occurs",
    is_flag=True,
)
@click.option(
    "--debug-log-file",
    help="Path of filename to save DEBUG log messages to",
    type=click.Path(
        writable=True,
    ),
)
@click.option(
    "-a",
    "--audit-log-file",
    help="Path of filename to save audit log to",
    type=click.Path(
        writable=True,
    ),
)
@click.option(
    "--audit-log-style",
    help="Style to use for audit log entries",
    type=click.Choice([style.name for style in AuditLog.Style]),
    default=AuditLog.Style.PLAIN.name,
)
@click.option(
    "-b",
    "--batch-size",
    help="How many deletes to process in one batch",
    type=click.INT,
)
@click.option(
    "-l",
    "--limit",
    help="Maximum number of of deletes to process",
    type=click.INT,
)
def start(
    config_file: str,
    dry_run: bool,
    debug_log_file: Optional[str],
    audit_log_file: Optional[str],
    audit_log_style: Optional[str],
    batch_size: Optional[int],
    limit: Optional[int],
) -> None:
    """Main entry point for app."""

    # pylint: disable=too-many-arguments

    asyncio.run(
        main(
            config_file=config_file,
            is_dry_run=dry_run,
            debug_log_file=debug_log_file,
            audit_log_file=audit_log_file,
            audit_log_style=audit_log_style,
            batch_size=batch_size,
            limit=limit,
        )
    )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['mastodonamnesia']

package_data = \
{'': ['*']}

install_requires = \
['aiocsv>=1.2.3,<2.0.0',
 'aiofiles>=0.8.0,<0.9.0',
 'arrow>=1.2.1,<2.0.0',
 'click>=8.1.3,<9.0.0',
 'minimal-activitypub>=0.4.0,<0.5.0',
 'outdated>=0.2.1,<0.3.0',
 'rich>=13.0.0,<14.0.0',
 'tqdm>=4.64.0,<5.0.0',
 'typing-extensions>=4.2.0,<5.0.0']

entry_points = \
{'console_scripts': ['mastodonamnesia = '
                     'mastodonamnesia.mastodon_amnesia:start']}

setup_kwargs = {
    'name': 'mastodonamnesia',
    'version': '2.2.1',
    'description': 'deletes old Mastodon toots',
    'long_description': '""""""""""""""""""""""""""\nMastodonAmnesia\n""""""""""""""""""""""""""\n\n|Repo| |CI| |Downloads|\n\n|Checked against| |Checked with| |Interrogate|\n\n|Code style| |Version| |Wheel|\n\n|AGPL|\n\n\n***!!! BEWARE, THIS TOOL WILL DELETE SOME OF YOUR MASTODON TOOTS !!!***\n\nMastodonAmnesia is a command line (CLI) tool to delete old toots from Mastodon or Pleroma instances.\nIt respects rate limits imposed by servers.\n\nInstall and run from `PyPi <https://pypi.org>`_\n=================================================\n\nIt\'s ease to install MastodonAmnesia from Pypi using the following command::\n\n    pip install mastodon_amnesia\n\nOnce installed MastodonManesia can be started by typing ``mastodonamnesia`` into the command line.\n\nInstall and run from `Source <https://codeberg.org/MarvinsMastodonTools/mastodonamnesia>`_\n==============================================================================================\n\nAlternatively you can run MastodonAmnesia from source by cloning the repository using the following command line::\n\n    git clone https://codeberg.org/MarvinsMastodonTools/mastodonamnesia.git\n\nMastodonAmnesia uses `Poetry <https://python-poetry.org/>`_ for dependency control, please install Poetry before proceeding further.\n\nBefore running, make sure you have all required python modules installed. With Poetry this is as easy as::\n\n    poetry install --no-dev\n\nRun MastodonAmnesia with the command `poetry run mastodonamnesia`\n\nConfiguration / First Run\n=========================\n\nMastodonAmnesia will ask for all necessary parameters when run for the first time and store them in ```config.json``\nfile in the current directory.\n\nLicensing\n=========\nMastodonAmnesia is licensed under the `GNU Affero General Public License v3.0 <http://www.gnu.org/licenses/agpl-3.0.html>`_\n\nSupporting MastodonAmnesia\n==========================\n\nThere are a number of ways you can support MastodonAmnesia:\n\n- Create an issue with problems or ideas you have with/for MastodonAmnesia\n- You can `buy me a coffee <https://www.buymeacoffee.com/marvin8>`_.\n- You can send me small change in Monero to the address below:\n\nMonero donation address\n-----------------------\n``86ZnRsiFqiDaP2aE3MPHCEhFGTeipdQGJZ1FNnjCb7s9Gax6ZNgKTyUPmb21WmT1tk8FgM7cQSD5K7kRtSAt1y7G3Vp98nT``\n\n\n.. |AGPL| image:: https://www.gnu.org/graphics/agplv3-with-text-162x68.png\n    :alt: AGLP 3 or later\n    :target:  https://codeberg.org/MarvinsMastodonTools/mastodonamnesia/src/branch/main/LICENSE.md\n\n.. |Repo| image:: https://img.shields.io/badge/repo-Codeberg.org-blue\n    :alt: Repo at Codeberg.org\n    :target: https://codeberg.org/MarvinsMastodonTools/mastodonamnesia\n\n.. |Downloads| image:: https://pepy.tech/badge/mastodonamnesia\n    :alt: Download count\n    :target: https://pepy.tech/project/mastodonamnesia\n\n.. |Code style| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :alt: Code Style: Black\n    :target: https://github.com/psf/black\n\n.. |Checked against| image:: https://img.shields.io/badge/Safety--DB-Checked-green\n    :alt: Checked against Safety DB\n    :target: https://pyup.io/safety/\n\n.. |Checked with| image:: https://img.shields.io/badge/pip--audit-Checked-green\n    :alt: Checked with pip-audit\n    :target: https://pypi.org/project/pip-audit/\n\n.. |Version| image:: https://img.shields.io/pypi/pyversions/mastodonamnesia\n    :alt: PyPI - Python Version\n\n.. |Wheel| image:: https://img.shields.io/pypi/wheel/mastodonamnesia\n    :alt: PyPI - Wheel\n\n.. |CI| image:: https://ci.codeberg.org/api/badges/MarvinsMastodonTools/mastodonamnesia/status.svg\n    :alt: CI / Woodpecker\n    :target: https://ci.codeberg.org/MarvinsMastodonTools/mastodonamnesia\n\n.. |Interrogate| image:: https://codeberg.org/MarvinsMastodonTools/mastodonamnesia/raw/branch/main/interrogate_badge.svg\n    :alt: Doc-string coverage\n    :target: https://interrogate.readthedocs.io/en/latest/\n',
    'author': 'marvin8',
    'author_email': 'marvin8@tuta.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://codeberg.org/MarvinsMastodonTools/mastodonamnesia',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""Package 'mastodonamnesia' level definitions."""
import sys
from typing import Any

from typing_extensions import Final


__version__: Final[str] = "2.2.1"
__display_name__: Final[str] = "MastodonAmnesia"
__package_name__: Final[str] = __display_name__.lower()
USER_AGENT: Final[
    str
] = f"{__display_name__}_v{__version__}_Python_{sys.version.split()[0]}"

Toot = dict[str, Any]
